import optuna as op
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder, LabelEncoder, StandardScaler, MinMaxScaler, MaxAbsScaler, RobustScaler
from category_encoders import WOEEncoder, TargetEncoder
from sklearn.impute import SimpleImputer
from imblearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer

def _instantiate_numerical_imputer(trial):
    strategy = trial.suggest_categorical(
        # 'most_frequent', 'constant'
        'numerical_imputer_strategy', ['mean', 'median']
    )
    return SimpleImputer(strategy=strategy, fill_value=-1)

def _instantiate_categorical_imputer(trial):
    strategy = trial.suggest_categorical(
        'categorical_imputer_strategy', ['most_frequent', 'constant']
    )
    return SimpleImputer(strategy=strategy, fill_value='missing')

def _instantiate_standard_scaler(trial):
    params = {
        # 'with_mean': trial.suggest_categorical('with_mean', [True, False]),
        # 'with_std': trial.suggest_categorical('with_std', [True, False])
    }
    return StandardScaler(**params)

def _instantiate_min_max_scaler(trial):
    params = {
        # 'clip': trial.suggest_categorical('clip', [True, False])
    }
    return MinMaxScaler(**params)

def _instantiate_max_abs_scaler(trial):
    return MaxAbsScaler()

def _instantiate_robust_scaler(trial):
    params = {
        # 'with_centering': trial.suggest_categorical('with_centering', [True, False]),
        # 'with_scaling': trial.suggest_categorical('with_scaling', [True, False]),
        # 'unit_variance': trial.suggest_categorical('unit_variance', [True, False])
    }
    return RobustScaler(**params)

def _instantiate_scaler(trial):
    scaler = trial.suggest_categorical(
        # 'maxabs', 'robust'
        'scaler', ['standard', 'minmax']
    )
    if scaler == 'standard':
        return _instantiate_standard_scaler(trial)
    elif scaler == 'minmax':
        return _instantiate_min_max_scaler(trial)
    elif scaler == 'maxabs':
        return _instantiate_max_abs_scaler(trial)
    elif scaler == 'robust':
        return _instantiate_max_abs_scaler(trial)

def _instantiate_one_hot_encoder(trial):
    return OneHotEncoder(drop='first')

def _instantiate_ordinal_encoder(trial):
    return OrdinalEncoder()

def _instantiate_target_encoder(trial):
    return TargetEncoder()

def _instantiate_woe_encoder(trial):
    return WOEEncoder()

def _instantiate_encoder(trial):
    encoder = trial.suggest_categorical(
        # 'target', 'woe'
        'encoder', ['onehot', 'ordinal']
    )
    if encoder == 'onehot':
        return _instantiate_one_hot_encoder(trial)
    elif encoder == 'ordinal':
        return _instantiate_ordinal_encoder(trial)
    # elif encoder == 'target':
    #     return _instantiate_target_encoder(trial)
    # elif encoder == 'woe':
    #     return _instantiate_woe_encoder(trial)

def _instantiate_numerical_pipeline(trial):
    pipeline = Pipeline([
        ('numerical_imputer', _instantiate_numerical_imputer(trial)),
        ('scaler', _instantiate_scaler(trial))
    ])
    return pipeline

def _instantiate_categorical_pipeline(trial):
    pipeline = Pipeline([
        ('categorical_imputer', _instantiate_categorical_imputer(trial)),
        ('encoder', _instantiate_encoder(trial))
    ])
    return pipeline

def _feature_selector(trial, features):
    select = lambda feature: trial.suggest_categorical(feature, [True, False])
    selected = [*filter(select, features)]
    return selected

def _instantiate_processor(trial, numerical_features, categorical_features, select=False):
    numerical_pipeline = _instantiate_numerical_pipeline(trial)
    categorical_pipeline = _instantiate_categorical_pipeline(trial)
    if select:
        numerical_features = _feature_selector(trial, numerical_features)
        numerical_features = _feature_selector(trial, categorical_features)
    processor = ColumnTransformer([
        ('numerical_pipeline', numerical_pipeline, numerical_features),
        ('categorical_pipeline', categorical_pipeline, categorical_features)
    ])
    return processor



import numpy as np
import pandas as pd
import optuna as op
import shap as sp
from sklearn.pipeline import Pipeline
from sklearn.model_selection import cross_val_score
from sklearn.metrics import accuracy_score, average_precision_score, recall_score, precision_score, f1_score, roc_auc_score, confusion_matrix
from ._processor import _instantiate_processor
from ._classifier import _instantiate_estimator


class ClassificationStudy:

    def __init__(self):
        self._study = None
        self._best_pipeline = None
        self._best_score = None

    def _instantiate_pipeline(trial, numerical_features, categorical_features, select=False):
        processor = _instantiate_processor(
            trial, numerical_features, categorical_features, select
        )
        estimator = _instantiate_estimator(trial)
        pipeline = Pipeline([
            ('processor', processor),
            ('estimator', estimator)
        ])
        return pipeline

    def optimize(self, X, y, metric='accuracy', select=False, study_name='optimization', cv=5, n_trials=100):

        numerical_features = [
            *X.select_dtypes(exclude=['object', 'category']).columns
        ]
        categorical_features = [
            *X.select_dtypes(include=['object', 'object']).columns
        ]

        def objective(self, trial):
            pipeline = self._instantiate_pipeline(
                trial, numerical_features, categorical_features, select
            )
            scores = cross_val_score(pipeline, X, y, scoring=metric, cv=cv)
            return np.min([np.mean(scores), np.median([scores])])

        op.logging.set_verbosity(op.logging.WARNING)
        self._study = op.create_study(study_name=study_name, direction='maximize')
        self._study.optimize(lambda trial: objective(self, trial, X, y), n_trials=n_trials)
        self._best_pipeline = self._instantiate_pipeline(self._study.best_trial, numerical_features, categorical_features, select)

    def fit(self, X, y):
        self._best_pipeline.fit(X, y)

    def evaluate(self, X, y, threshold=0.5):
        pred_y_proba = self._best_pipeline.predict_proba(X)[:, 1]
        scores = {
            'AUC-ROC': roc_auc_score(y, pred_y_proba),
            'Accuracy': accuracy_score(y, pred_y_proba >= threshold),
            'Average Precision': average_precision_score(y, pred_y_proba),
            'Recall': recall_score(y, pred_y_proba >= threshold),
            'Precision': precision_score(y, pred_y_proba >= threshold),
            'F1': f1_score(y, pred_y_proba >= threshold)
        }
        return pred_y_proba, pd.DataFrame(scores).iloc[0,:].to_frame().T.round(4)

    def plot_optimization_history(self):
        return op.visualization.plot_optimization_history(self._study)

    def plot_param_importances(self):
        return op.visualization.plot_param_importances(self._study)

    def plot_feature_importances(self, X):
        explainer = sp.Explainer(self._best_pipeline[-1])
        X_transformed = self._best_pipeline[0].transform(X)
        feature_names = [feature.split('__')[1] for feature in self._best_pipeline[0].get_feature_names_out()]
        shap_values = explainer.shap_values(X_transformed)
        return sp.summary_plot(shap_values, X_transformed, feature_names)
    
    @property
    def best_pipeline(self):
        return self._best_pipeline

    @property
    def best_score(self):
        return self._best_score


from setuptools import find_packages, setup
from typing import List

def get_requirements(file_path):
    requirements = []
    with open(file_path) as file:
        requirements = file.readlines()
        requirements = [req.replace('\n', '') for req in requirements]
        if '-e .' in requirements:
            requirements.remove('-e .')
    return requirements

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name = 'baseline_optimal',
    version = '0.0.3',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author = 'Shunji Wan',
    author_email = 'sw3843@columbia.edu',
    url='https://github.com/sjwan01/baseline_optimal',
    packages = find_packages(),
    install_requires = get_requirements('requirements.txt')
)


import optuna as op
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from xgboost import XGBClassifier

def _instantiate_estimator(trial):
    estimator = trial.suggest_categorical(
        'estimator', ['decision_tree', 'random_forest', 'adaboost', 'xgboost']
    )
    if estimator == 'decision_tree':
        return _instantiate_decision_tree(trial)
    elif estimator == 'random_forest':
        return _instantiate_random_forest(trial)
    elif estimator == 'adaboost':
        return _instantiate_adaboost(trial)
    elif estimator == 'xgboost':
        return _instantiate_xgboost(trial)

def _instantiate_decision_tree(trial):
    params = {
        'max_features': trial.suggest_categorical('dt_max_features', ['log2', 'sqrt', None]),
        'max_depth': trial.suggest_int('dt_max_depth', 5, 10),
        'min_samples_split': trial.suggest_int('dt_min_sample_split', 2, 10),
        'random_state': 123
    }
    return DecisionTreeClassifier(**params)

def _instantiate_random_forest(trial):
    params = {
        'n_estimators': trial.suggest_int('rf_n_estimators', 100, 300),
        'max_features': trial.suggest_categorical('rf_max_features', ['log2', 'sqrt', None]),
        'max_depth': trial.suggest_int('rf_max_depth', 5, 10),
        'min_samples_split': trial.suggest_int('rf_min_sample_split', 2, 10),
        'random_state': 123
    }
    return RandomForestClassifier(**params)

def _instantiate_adaboost(trial):
    params = {
        'n_estimators': trial.suggest_int('ada_n_estimators', 100, 300),
        'learning_rate': trial.suggest_float('ada_learning_rate', 0.001, 0.1),
        'random_state': 123
    }
    return AdaBoostClassifier(**params)

def _instantiate_xgboost(trial):
    params = {
        'eta': trial.suggest_float('xgb_eta', 0, 0.1),
        'n_estimators': trial.suggest_int('xgb_n_estimators', 100, 300),
        'max_features': trial.suggest_categorical('xgb_max_features', ['log2', 'sqrt', None]),
        'max_depth': trial.suggest_int('xgb_max_depth', 5, 10),
        'min_samples_split': trial.suggest_int('xgb_min_sample_split', 2, 10),
        'random_state': 123
    }
    return XGBClassifier(**params)